"""Types collection"""

import asyncio
import enum

__all__ = [
    "DisplayEventName",
    "DisplayEvent",
]


class DisplayEventName(enum.Enum):
    """Valid event names"""

    ON_ACTION_MESSAGE = "on_action_message"
    ON_ACTION_ERROR = "on_action_error"
    ON_RUNNER_START = "on_runner_start"
    ON_RUNNER_FINISH = "on_runner_finish"
    ON_PLAN_INTERACTION = "on_plan_interaction"
    ON_ACTION_START = "on_action_start"
    ON_ACTION_FINISH = "on_action_finish"


class DisplayEvent:
    """Display-related event floating to the runner"""

    def __init__(self, name: DisplayEventName, **kwargs) -> None:
        self.name: DisplayEventName = name
        self.kwargs: dict = kwargs
        self.future: asyncio.Future = asyncio.Future()



"""A convenient task runner"""

from .actions.base import (
    ActionBase,
    ArgsBase,
    EmissionScannerActionBase,
)
from .actions.types import Stderr
from .config.constants import C
from .loader.default import DefaultYAMLWorkflowLoader
from .runner import Runner
from .strategy.impl import (
    FreeStrategy,
    SequentialStrategy,
    ExplicitStrategy,
)
from .version import __version__



"""Runner output processor default"""

import typing as t

from . import dialog
from .base import BaseDisplay
from .color import Color
from .utils import Tree, locate_parent_name_by_prefix
from ..actions.types import (
    Stderr,
    NamedMessageSource,
    RenamedMessageSource,
    ActionStatus,
)
from ..exceptions import InteractionError
from ..workflow import Workflow

__all__ = [
    "PrologueDisplay",
    "HeaderDisplay",
    "PrefixDisplay",
    "KNOWN_DISPLAYS",
]

ColorWrapperType = t.Callable[[str], str]


class PrologueDisplay(BaseDisplay):
    """Default display base"""

    NAME: str
    STATUS_TO_MARK_SYMBOL_MAP: dict[ActionStatus, str] = {
        ActionStatus.SKIPPED: "◯",
        ActionStatus.PENDING: "◯",
        ActionStatus.FAILURE: "✗",
        ActionStatus.WARNING: "✓",
        ActionStatus.RUNNING: "◯",
        ActionStatus.SUCCESS: "✓",
        ActionStatus.OMITTED: "◯",
    }
    STATUS_TO_COLOR_WRAPPER_MAP: dict[ActionStatus, ColorWrapperType] = {
        ActionStatus.SKIPPED: Color.gray,
        ActionStatus.PENDING: Color.gray,
        ActionStatus.FAILURE: Color.red,
        ActionStatus.WARNING: Color.yellow,
        ActionStatus.RUNNING: lambda x: x,
        ActionStatus.SUCCESS: Color.green,
        ActionStatus.OMITTED: Color.gray,
    }

    def __init__(self) -> None:
        super().__init__()
        self._status_topology: Tree[NamedMessageSource] = Tree()
        self._last_displayed_name: t.Optional[str] = None

    def _make_prologue(self, source: NamedMessageSource, mark: str) -> str:
        raise NotImplementedError

    def on_runner_start(self, children: t.Iterable[NamedMessageSource]) -> None:
        if not self._status_topology:
            self._status_topology.put((action.name, action) for action in children)
            return
        children_list: list[NamedMessageSource] = list(children)
        corresponding_action_name = locate_parent_name_by_prefix(
            children=(action.name for action in children_list),
            candidates=self._status_topology,
        )
        longest_match_length = len(corresponding_action_name)
        self._status_topology.put(
            (
                (action.name, RenamedMessageSource(origin=action, name=action.name[longest_match_length + 1 :]))
                for action in children_list
            ),
            parent_name=corresponding_action_name,
        )

    def on_action_message(self, source: NamedMessageSource, message: str) -> None:
        is_stderr: bool = isinstance(message, Stderr)
        for line in message.splitlines() if message else [message]:
            line_prefix: str = self._make_prologue(source=source, mark="*" if is_stderr else " ")
            self.display(f"{line_prefix}{Color.yellow(line) if is_stderr else line}")

    def on_action_error(self, source: NamedMessageSource, message: str) -> None:
        line_prefix: str = self._make_prologue(source=source, mark="!")
        for line in message.splitlines():
            super().on_action_error(
                source=source,
                message=f"{line_prefix}{Color.red(line)}",
            )

    def _generate_status_banner_lines(self) -> t.Generator[str, None, None]:
        for ascii_tree_prefix, source in self._status_topology.generate_ascii_representation():
            color: ColorWrapperType = self.STATUS_TO_COLOR_WRAPPER_MAP[source.status]
            status_mark: str = self.STATUS_TO_MARK_SYMBOL_MAP[source.status]
            status_part: str = f"{status_mark} {source.status.value}"
            yield f"{color(status_part)}: {Color.gray(ascii_tree_prefix)}{color(source.name)}"

    def on_runner_finish(self) -> None:
        """Show a text banner with the status info"""
        for line in self._generate_status_banner_lines():
            self.display(line)

    def on_plan_interaction(self, workflow: Workflow) -> None:
        displayed_action_names_with_descriptions: list[t.Tuple[str, str]] = []
        default_selected_action_names: list[str] = []
        for action in workflow.iterate_actions():
            if action.selectable:
                action_name_with_description: str = action.name
                if action.description is not None:
                    action_name_with_description = f"{action_name_with_description}: {action.description}"
                displayed_action_names_with_descriptions.append((action_name_with_description, action.name))
                default_selected_action_names.append(action.name)
        if not displayed_action_names_with_descriptions:
            raise InteractionError("No selectable actions found")
        selected_action_names: list[str] = dialog.run_dialog(
            choices=displayed_action_names_with_descriptions,
            default=default_selected_action_names,
        )
        self.logger.warning(f"Interactively selected actions: {selected_action_names}")
        for action in workflow.iterate_actions():
            if action.name in default_selected_action_names and action.name not in selected_action_names:
                self.logger.info(f"Disabling action: {action.name}")
                action.enabled = False


class PrefixDisplay(PrologueDisplay):
    """Adds prefixes to output chunks"""

    NAME = "prefixes"

    def __init__(self) -> None:
        super().__init__()
        self._action_names_max_len: int = 0

    def on_runner_start(self, children: t.Iterable[NamedMessageSource]) -> None:
        super().on_runner_start(children)
        self._action_names_max_len = max(self._action_names_max_len, *map(len, self._status_topology))

    def _make_prologue(self, source: NamedMessageSource, mark: str) -> str:
        """Construct prefix based on previous emitter action name"""
        justification_len: int = self._action_names_max_len + 2  # "2" here stands for square brackets
        formatted_name: str = (
            f"[{source.name}]".ljust(justification_len)
            if self._last_displayed_name != source.name
            else " " * justification_len
        )
        self._last_displayed_name = source.name
        return Color.gray(f"{formatted_name} {mark}| ")


class HeaderDisplay(PrologueDisplay):
    """Adds headers to output chunks"""

    NAME = "headers"

    def _close_block_if_necessary(self) -> None:
        if self._last_displayed_name is not None:
            self.display(Color.gray(" ╵"))

    def _make_prologue(self, source: NamedMessageSource, mark: str) -> str:
        """Construct header based on previous emitter action name"""
        if self._last_displayed_name != source.name:
            self._close_block_if_necessary()
            self.display(Color.gray(f" ┌─[{source.name}]"))
            self._last_displayed_name = source.name
        return Color.gray(f"{mark}│ ")

    def _generate_status_banner_lines(self) -> t.Generator[str, None, None]:
        """Add extra space in the beginning of the message, so it is aligned with the stream"""
        for line in super()._generate_status_banner_lines():
            yield f" {line}"

    def on_runner_finish(self) -> None:
        self._close_block_if_necessary()
        super().on_runner_finish()


KNOWN_DISPLAYS: dict[str, type[BaseDisplay]] = {
    HeaderDisplay.NAME: HeaderDisplay,
    PrefixDisplay.NAME: PrefixDisplay,
}



"""All the templating stuff."""

import io
import tokenize
import typing as t

__all__ = [
    "ExpressionTokenizer",
    "TemplarStringLexer",
]


class ExpressionTokenizer:
    """Utilize tokenize.tokenize while simultaneously tracking the caret position"""

    def __init__(self, data: str) -> None:
        self._bytes: bytes = data.encode()
        self._stream: io.BytesIO = io.BytesIO(self._bytes)
        self._token_generator: t.Iterator[tokenize.TokenInfo] = self._careless_tokenize()
        self._scanned_lines_length_sum: int = 0
        self._prev_line_length: int = 0
        self.position: int = 0

    def _readline(self) -> bytes:
        """Memorize read lines length sum"""
        line: bytes = self._stream.readline()
        self._scanned_lines_length_sum += self._prev_line_length
        self._prev_line_length = len(line)
        return line

    def _careless_tokenize(self) -> t.Generator[tokenize.TokenInfo, None, None]:
        """Tokenize the stream, while ignoring all `TokenError`s"""
        try:
            yield from tokenize.tokenize(self._readline)
        except tokenize.TokenError:
            pass

    def get_token(self) -> tokenize.TokenInfo:
        """Yield a token and memorize its original position"""
        token: tokenize.TokenInfo = next(self._token_generator)
        self.position = self._scanned_lines_length_sum + token.start[1]
        return token

    def __iter__(self) -> t.Iterator[tokenize.TokenInfo]:
        return self

    def __next__(self) -> tokenize.TokenInfo:
        return self.get_token()


class TemplarStringLexer:
    """Emit raw text and expressions separately"""

    TEXT: int = 0
    EXPRESSION: int = 1
    _IGNORED_TOKENS_TYPES = [
        tokenize.NL,
        tokenize.NEWLINE,
        tokenize.ENCODING,
        tokenize.INDENT,
        tokenize.DEDENT,
        tokenize.ENDMARKER,
    ]

    def __init__(self, data: str) -> None:
        self._data: str = data
        self._len: int = len(data)
        self._caret: int = 0

    def _get_symbol(self) -> str:
        """Read the next symbol of the input data"""
        if self._caret >= self._len:
            raise EOFError
        self._caret += 1
        return self._data[self._caret - 1]

    def __iter__(self) -> t.Iterator[t.Tuple[int, str]]:
        """Alternately yield raw text to leave as is and expressions to evaluate"""
        armed_at: bool = False
        text_start: int = self._caret
        while True:
            try:
                # Each expression starts with an "@{" pair, which can be escaped by doubling the "@" down,
                # thus each second "@" disengages the expression scanning readiness
                if (symbol := self._get_symbol()) == "@":
                    armed_at = not armed_at
                    continue
                if symbol == "{" and armed_at:
                    if maybe_text := self._data[text_start : self._caret - 2]:  # -2 stands for the "@{"
                        yield self.TEXT, maybe_text
                    expression_source_length, expression = self._read_expression()
                    yield self.EXPRESSION, expression
                    text_start = self._caret + expression_source_length + 1  # Start again right after the closing brace
                armed_at = False
            except (StopIteration, EOFError):
                if maybe_text := self._data[text_start:]:
                    yield self.TEXT, maybe_text
                break

    def _read_expression(self) -> t.Tuple[int, str]:
        """Use a tokenizer to detect the closing brace"""
        brace_depth: int = 0
        collected_tokens: list[t.Tuple[int, str]] = []
        tokenizer = ExpressionTokenizer(data=self._data[self._caret :])
        while True:
            token_info = tokenizer.get_token()
            if token_info.exact_type == tokenize.LBRACE:
                brace_depth += 1
            elif token_info.exact_type == tokenize.RBRACE:
                brace_depth -= 1
                if brace_depth < 0:
                    clean_expression: str = tokenize.untokenize(collected_tokens)
                    return tokenizer.position, clean_expression
            if token_info.exact_type not in self._IGNORED_TOKENS_TYPES:
                collected_tokens.append((token_info.type, token_info.string))



"""Runner output processor base"""

import typing as t

from ..actions.types import NamedMessageSource
from ..exceptions import InteractionError
from ..logging import WithLogger
from ..workflow import Workflow

__all__ = [
    "BaseDisplay",
]


class BaseDisplay(WithLogger):
    """Base class for possible customizations"""

    def display(self, message: str) -> None:
        """Send text to the end user"""
        print(message.rstrip("\n"))

    # pylint: disable=unused-argument
    def on_action_message(self, source: NamedMessageSource, message: str) -> None:
        """Process a message from some source"""
        self.display(message)  # pragma: no cover

    # pylint: disable=unused-argument
    def on_action_error(self, source: NamedMessageSource, message: str) -> None:
        """Process an error from some source"""
        self.display(message)  # pragma: no cover

    def on_runner_start(self, children: t.Iterable[NamedMessageSource]) -> None:
        """Runner start callback"""

    def on_runner_finish(self) -> None:
        """Runner finish callback"""

    def on_plan_interaction(self, workflow: Workflow) -> None:
        """Execution plan approval callback"""
        raise InteractionError  # pragma: no cover

    def on_action_start(self, source: NamedMessageSource) -> None:
        """Action start callback"""

    def on_action_finish(self, source: NamedMessageSource) -> None:
        """Action finish callback"""



"""Available execution strategies"""

from __future__ import annotations

import asyncio
import typing as t

from .base import BaseStrategy
from ..actions.base import WorkflowActionExecution, ActionDependency
from ..exceptions import PendingActionUnresolvedOutcomeError, AutoStrategyCycleError
from ..workflow import Workflow

__all__ = [
    "ExplicitStrategy",
    "FreeStrategy",
    "SequentialStrategy",
    "AutoStrategy",
]


class ExplicitStrategy(BaseStrategy):
    """Actions are started immediately after their explicit dependencies have finished the execution.
    If no dependencies given for an action, then it is scheduled to start in the very beginning of the workflow run."""

    NAME = "explicit"

    def __init__(self, workflow: Workflow) -> None:
        super().__init__(workflow)
        self._pending: set[WorkflowActionExecution] = set(self._workflow.values())
        self._running: set[WorkflowActionExecution] = set()

    def _find_schedulable_execution(self) -> t.Optional[WorkflowActionExecution]:
        """Completely non-optimal (always scan all actions), but readable yet"""
        for execution in self._pending | self._running:
            if execution.future.done():
                self._running.discard(execution)
                continue
            if execution in self._pending and all(
                self._workflow[dependency.name].future.done() for dependency in execution.ancestors
            ):
                self.logger.debug(f"Action {execution.name!r} is ready for scheduling")
                return execution
        return None

    async def __anext__(self) -> WorkflowActionExecution:
        while True:
            # Get an execution and check whether to emit or to skip it
            execution: WorkflowActionExecution = await self._next_execution()
            self.logger.debug(f"The next action is: {execution}")
            self._pending.remove(execution)
            self._running.add(execution)
            return execution

    async def _next_execution(self) -> WorkflowActionExecution:
        while True:
            # Do we have anything pending already?
            if maybe_next_execution := self._find_schedulable_execution():
                return maybe_next_execution
            if not self._running:
                raise StopAsyncIteration
            await asyncio.wait(
                [execution.future for execution in self._running],
                return_when=asyncio.FIRST_COMPLETED,
            )


class FreeStrategy(ExplicitStrategy):
    """All actions are started immediately in parallel. All dependencies are ignored."""

    NAME = "free"

    def __init__(self, workflow: Workflow) -> None:
        super().__init__(workflow)
        for execution in self._workflow.values():
            if execution.ancestors:
                self.logger.warning(f"Ignoring action {execution.name!r} dependencies: {execution.ancestors}")
                execution.ancestors.clear()


class SequentialStrategy(FreeStrategy):
    """Actions run one-by-one in the same order they are specified in the workflow. All dependencies are ignored."""

    NAME = "sequential"

    def __init__(self, workflow: Workflow) -> None:
        super().__init__(workflow)
        executions: list[WorkflowActionExecution] = list(self._workflow.values())
        for num, execution in enumerate(executions):
            if num > 0:
                ancestor_name: str = executions[num - 1].name
                self.logger.debug(f"Adding an implicit dependency of {execution.name!r} on {ancestor_name!r}")
                execution.ancestors.append(ActionDependency(name=ancestor_name))


class AutoStrategy(ExplicitStrategy):
    """Use both explicit and outcome-based dependencies. Same as [](#explicit),
    but in case of referring to some action's outcome,
    an implicit dependency is added so the outcome could be resolved."""

    NAME = "auto"

    async def _next_execution(self) -> WorkflowActionExecution:
        try:
            while True:
                execution = await super()._next_execution()
                try:
                    execution.render_action_args()
                except PendingActionUnresolvedOutcomeError as e:
                    self.logger.info(f"Adding an automatic dependency of {execution.name!r} on {e.action_name!r}")
                    execution.ancestors.append(ActionDependency(name=e.action_name))
                    continue
                except Exception:  # nosec
                    # All other rendering exceptions are not important at this point
                    pass
                return execution
        except StopAsyncIteration:
            if self._pending:
                pending_actions: str = ", ".join(sorted(repr(execution.name) for execution in self._pending))
                error_text: str = (
                    f"The following actions went unreachable due to implicit dependencies: "
                    f"{pending_actions}. Check outcome references for circularity."
                )
                raise AutoStrategyCycleError(error_text) from None
            raise



"""Common loader utilities"""

from __future__ import annotations

import sys
import types
import typing as t

__all__ = [
    "get_class_annotations",
]

OptionalDict = t.Optional[dict[str, t.Any]]


def get_class_annotations(obj: t.Any) -> dict[str, t.Any]:
    """Different pythons hold __annotations__ attribute values as strings or real types, depending on the version."""
    # Resolve class first
    class_definition: type = obj if isinstance(obj, type) else obj.__class__
    class_dict: OptionalDict = getattr(class_definition, "__dict__", None)
    annotations_source: OptionalDict = None
    if class_dict is not None and hasattr(class_dict, "get"):
        annotations_source = class_dict.get("__annotations__", None)
        if isinstance(annotations_source, types.GetSetDescriptorType):
            annotations_source = None  # pragma: no cover

    class_globals: OptionalDict = None
    module_name: t.Optional[str] = getattr(class_definition, "__module__", None)
    if module_name:
        module: t.Optional[types.ModuleType] = sys.modules.get(module_name, None)
        if module:
            class_globals = getattr(module, "__dict__", None)
    class_locals = dict(vars(class_definition))

    if annotations_source is None:
        return {}

    if not isinstance(annotations_source, dict):
        raise ValueError(f"{class_definition!r}.__annotations__ is neither a dict nor None")  # pragma: no cover

    # Unwrap class
    while True:
        if hasattr(class_definition, "__wrapped__"):  # pragma: no cover
            class_definition = class_definition.__wrapped__
            continue
        break
    if hasattr(class_definition, "__globals__"):
        class_globals = class_definition.__globals__  # pragma: no cover

    # Eval finally
    return {
        # pylint: disable=eval-used
        key: value if not isinstance(value, str) else eval(value, class_globals, class_locals)  # nosec
        for key, value in annotations_source.items()
    }



"""Command-line interface entry"""

import functools
import sys
import typing as t
from logging import getLogger

import click

from . import logging as grana_logging
from .config.constants import C, rc
from .config.constants.base import ConstantSource
from .config.constants.cli import get_cli_option, cli_opts_receiver
from .display.color import Color
from .exceptions import BaseError, ExecutionFailed
from .loader.default import DefaultYAMLWorkflowLoader
from .runner import Runner
from .tools.proxy import DeferredCallsProxy
from .version import __version__

logger = DeferredCallsProxy(obj=getLogger(__name__))


class WorkflowPositionalArgument(click.Argument):
    """Optional positional argument for the workflow source"""

    # pylint: disable=unused-argument
    def __init__(self, param_decls: t.Sequence[str], required: t.Optional[bool] = None, **attrs: t.Any) -> None:
        super().__init__(param_decls, required=False, nargs=-1)

    def get_help_record(self, ctx: click.Context) -> t.Optional[t.Tuple[str, str]]:
        return self.make_metavar(), (
            "Workflow source file. When not given, will look for one of grana.yml/grana.yaml "
            "files in the context directory. Use the '-' value to read yaml configuration from the standard input"
        )

    def process_value(self, ctx: click.Context, value: t.Any) -> t.Optional[str]:
        """Check that there is not more than one argument for the workflow source"""
        vanilla_value: t.Tuple[str, ...] = super().process_value(ctx, value)
        if len(vanilla_value) > 1:
            raise click.BadParameter("Cannot apply more than one value", param=self)
        return vanilla_value[0] if vanilla_value else None

    def make_metavar(self) -> str:
        """Fixed representation"""
        return "[WORKFLOW_FILE]"


@click.group
@click.option(
    "-l",
    "--log-level",
    help="Logging subsystem level",
)
@click.option(
    "-L",
    "--log-file",
    help="Log file path",
)
@click.option(
    "-d",
    "--display",
    help="Display name",
)
@cli_opts_receiver
def main() -> None:
    """Open-source command-line task automation tool."""


def wrap_cli_command(func):
    """Standard loading and error handling"""

    @cli_opts_receiver
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        # Enable constants caches
        with C.mount_context_cache():
            grana_logging.configure_logging(
                main_file=C.LOG_FILE,
                level=C.LOG_LEVEL,
                colorize=C.USE_COLOR and not C.LOG_FILE,
            )
            logger.uncork()
            rc.logger.uncork()
            try:
                return func(*args, **kwargs)
            except BaseError as e:
                logger.debug("", exc_info=True)
                sys.stderr.write(f"! {e}\n")
                sys.exit(e.CODE)
            except ExecutionFailed:
                logger.debug("Some steps failed")
                sys.exit(1)
            except Exception as e:
                logger.debug("", exc_info=True)
                sys.stderr.write(f"! UNHANDLED EXCEPTION: {e!r}\n")
                sys.exit(2)

    return wrapped


@main.command
@wrap_cli_command
@click.option(
    "-s",
    "--strategy",
    help="Execution strategy for the workflow",
)
@click.option("-i", "--interactive", help="Run in dialog mode.", is_flag=True, default=False)
@click.argument("workflow_file", cls=WorkflowPositionalArgument, help="Workflow file path")
def run() -> None:
    """Run the pipeline."""
    Runner().run_sync()


@main.command
@wrap_cli_command
@click.argument("workflow_file", cls=WorkflowPositionalArgument)
def validate() -> None:
    """Check workflow source validity.
    Return code is zero, when validation passes."""
    action_num: int = len(Runner().workflow)
    logger.info(f"Located actions number: {action_num}")


@main.command
@wrap_cli_command
def version() -> None:
    """Display package version."""
    print(__version__)


@main.group
def info() -> None:
    """Miscellaneous tool information."""


@info.command
def env_vars() -> None:
    """Shows environment variables names that are taken into account."""
    print(C.env_doc())


@info.command
@wrap_cli_command
@click.option("--show-defaults", help="Show constants with default values", is_flag=True, default=False)
def runtime() -> None:
    """Shows runtime information."""
    display_spool: list[str] = []

    def section(name: str) -> None:
        display_spool.append(f"\n{Color.bold(name)}")

    def mapping(name: str) -> None:
        display_spool.append(f"{Color.yellow(name)}:")

    def kv(k: str, v: t.Any, *, indent: int = 0) -> None:
        display_spool.append(f"{'    ' * indent}{Color.blue(k)}: {Color.green(str(v))}")

    section("Python")
    kv("Version", sys.version.split(" ", 1)[0])
    kv("Executable", sys.executable)

    section("Configuration")
    for const_descriptor in C.constants_info():
        if const_descriptor.effective_source == ConstantSource.DEFAULT and not get_cli_option("show_defaults"):
            continue
        mapping(const_descriptor.name)
        kv("Value", const_descriptor.value, indent=1)
        kv("Source", const_descriptor.effective_source.name.lower(), indent=1)

    section("Actions")
    for actions_name, (action_class, action_source) in sorted(
        DefaultYAMLWorkflowLoader().get_action_factories_info().items()
    ):
        mapping(actions_name)
        if doc := getattr(action_class, "__doc__", ""):
            kv("Info", doc, indent=1)
        kv("Source", action_source, indent=1)

    d = C.INTERNAL_DISPLAY_CLASS()
    for line in display_spool:
        d.display(line)



"""
Runner has too many dependencies,
thus placed to a separate module.
"""

import asyncio
import functools
import io
import sys
import typing as t
from pathlib import Path

from . import types
from .actions.base import WorkflowActionExecution
from .actions.types import ActionStatus
from .config.constants import C
from .display.types import DisplayEvent, DisplayEventName
from .exceptions import SourceError, ExecutionFailed, ActionRenderError, ActionRunError
from .loader.helpers import get_default_loader_class_for_source
from .logging import WithLogger
from .workflow import Workflow

__all__ = [
    "Runner",
]

IOType = io.TextIOBase


class Runner(WithLogger):
    """Main entry object"""

    def __init__(self, source: t.Union[Path, IOType, dict, None] = None) -> None:
        self._workflow_source: t.Union[Path, IOType, dict]
        if source is not None:
            self._workflow_source = source
        else:
            self._workflow_source = self._detect_workflow_source()
        self._started: bool = False
        self._execution_failed: bool = False

    @functools.cached_property
    def _events_flow(self) -> asyncio.Queue:
        return asyncio.Queue()

    @functools.cached_property
    def loader(self) -> types.LoaderType:
        """Workflow loader"""
        loader_class: types.LoaderClassType
        if C.WORKFLOW_LOADER_CLASS:
            loader_class = C.WORKFLOW_LOADER_CLASS
        else:
            loader_class = get_default_loader_class_for_source(self._workflow_source)
        self.logger.debug(f"Using workflow loader class: {loader_class}")
        return loader_class()

    @functools.cached_property
    def workflow(self) -> Workflow:
        """Calculated workflow"""
        if isinstance(self._workflow_source, io.TextIOBase):
            return self.loader.load_from_text(self._workflow_source.read())
        if isinstance(self._workflow_source, dict):
            return self.loader.load_from_dict(self._workflow_source)
        return self.loader.load_from_file(self._workflow_source)

    @functools.cached_property
    def display(self) -> types.DisplayType:
        """Attached display"""
        display_class: types.DisplayClassType = C.EXTERNAL_DISPLAY_CLASS or C.INTERNAL_DISPLAY_CLASS
        self.logger.debug(f"Using display class: {display_class}")
        return display_class()

    @classmethod
    def _detect_workflow_source(cls) -> t.Union[Path, IOType]:
        if source_file := C.WORKFLOW_SOURCE_FILE:
            if str(source_file) == "-":
                cls.logger.info("Using stdin as workflow source")
                return t.cast(IOType, sys.stdin)
            if not source_file.exists():
                raise SourceError(f"Given workflow file does not exist: {source_file}")
            cls.logger.info(f"Using given workflow file: {source_file}")
            return source_file
        scan_path: Path = C.CONTEXT_DIRECTORY
        cls.logger.debug(f"Looking for workflow files at {str(scan_path)!r}")
        located_source_file: t.Optional[Path] = None
        for candidate_file_name in (
            "grana.yml",
            "grana.yaml",
        ):  # type: str
            if (maybe_source_file := scan_path / candidate_file_name).exists():
                cls.logger.info(f"Detected the workflow source: {str(maybe_source_file)!r}")
                if located_source_file is not None:
                    raise SourceError(f"Multiple workflow sources detected in {scan_path}")
                located_source_file = maybe_source_file
        if located_source_file is None:
            raise SourceError(f"No workflow source detected in {scan_path}")
        return located_source_file

    def _send_display_event(self, name: DisplayEventName, **kwargs) -> asyncio.Future:
        """Create a display event and return a future which indicates event processing status"""
        self._events_flow.put_nowait(event := DisplayEvent(name, **kwargs))
        return event.future

    async def _process_display_events(self) -> None:
        while True:
            event: DisplayEvent = await self._events_flow.get()
            try:
                display_method = getattr(self.display, event.name.value)
                display_method(**event.kwargs)
            except Exception as e:
                self.logger.exception(f"`{event.name}` callback failed for {self.display}")
                event.future.set_exception(e)
            else:
                event.future.set_result(None)

    async def run_async(self) -> None:
        """Primary coroutine for all further processing"""
        if self._started:
            raise RuntimeError("Runner has been started more than one time")
        self._started = True
        # Build workflow and display
        workflow: Workflow = self.workflow
        with workflow.configuration.apply():
            display: types.DisplayType = self.display
            display.logger.debug("Starting events processing")
            display_events_flow_processing_task: asyncio.Task = asyncio.create_task(self._process_display_events())
            try:
                await self._send_display_event(
                    DisplayEventName.ON_RUNNER_START,
                    children=workflow.iterate_actions(),
                )
                if C.INTERACTIVE_MODE:
                    await self._send_display_event(DisplayEventName.ON_PLAN_INTERACTION, workflow=workflow)
                await self._run_all_actions()
                await self._send_display_event(DisplayEventName.ON_RUNNER_FINISH)
                if self._execution_failed:
                    raise ExecutionFailed
            finally:
                display_events_flow_processing_task.cancel()

    async def _run_all_actions(self) -> None:
        action_runners: dict[WorkflowActionExecution, asyncio.Task] = {}
        strategy_class: types.StrategyClassType = C.STRATEGY_CLASS
        strategy: types.StrategyType = strategy_class(workflow=self.workflow)
        async for action in strategy:  # type: WorkflowActionExecution
            # Finalize all actions that have been done already
            for maybe_finished_action, corresponding_runner_task in list(action_runners.items()):
                if maybe_finished_action.future.done():
                    self.logger.debug(f"Finalizing done action {maybe_finished_action.name!r} runner")
                    await corresponding_runner_task
                    action_runners.pop(maybe_finished_action)
            self.logger.debug(f"Allocating action runner for {action.name!r}")
            action_runners[action] = asyncio.create_task(self._run_action(action=action))

        # Finalize running actions
        for task in action_runners.values():
            await task

    async def _dispatch_action_messages_to_display(self, action: WorkflowActionExecution) -> None:
        async for event in action.read_messages():
            self._events_flow.put_nowait(event)

    async def _run_action(self, action: WorkflowActionExecution) -> None:
        if not action.enabled:
            action.omit()
            return None
        for dependency in action.ancestors:
            ancestor: WorkflowActionExecution = self.workflow[dependency.name]
            if (
                ancestor.status in (ActionStatus.FAILURE, ActionStatus.SKIPPED, ActionStatus.WARNING)
                and dependency.strict
            ):
                self.logger.debug(f"Action {action} is qualified as skipped due to strict failure: {ancestor}")
                action.skip()
                return None
        self.logger.debug(f"Calling `{DisplayEventName.ON_ACTION_START}` for {action.name!r}")
        await self._send_display_event(DisplayEventName.ON_ACTION_START, source=action)
        self.logger.debug(f"Allocating action dispatcher for {action.name!r}")
        action_messages_reader_task: asyncio.Task = asyncio.create_task(
            self._dispatch_action_messages_to_display(action=action)
        )
        try:
            await action.execute()
        except Exception as e:
            message: str
            if isinstance(e, ActionRunError):
                message = str(e)
            elif isinstance(e, ActionRenderError):
                message = f"Action {action.name!r} rendering failed: {e}"
            else:
                message = f"Action {action.name!r} run exception: {e!r}"
            if message:
                await self._send_display_event(
                    DisplayEventName.ON_ACTION_ERROR,
                    source=action,
                    message=message,
                )
            if action.status == ActionStatus.WARNING:
                self.logger.warning(f"Action {action.name!r} finished with warning status")
            else:
                self.logger.warning(f"Action {action.name!r} execution failed: {e!r}")
                self._execution_failed = True
            self.logger.debug("Action failure traceback", exc_info=True)
        finally:
            await action_messages_reader_task
            self.logger.debug(f"Calling `{DisplayEventName.ON_ACTION_FINISH.value}` for {action.name!r}")
            await self._send_display_event(DisplayEventName.ON_ACTION_FINISH, source=action)

    def run_sync(self):
        """Wrap async run into an event loop"""
        asyncio.run(self.run_async())



"""Lazy-loaded constants base machinery"""

from __future__ import annotations

import dataclasses
import enum
import os
import typing as t
from pathlib import Path

from . import workflow, rc
from .cache import CACHE
from .cli import get_cli_option
from ...logging import WithLogger
from ...rendering.containers import LazyProxy

__all__ = [
    "Inapplicable",
    "ConstantSource",
    "ConstantBase",
    "ConstantProxyDescriptor",
    "ConstantBool",
    "ConstantPathList",
    "ConstantPath",
    "sentinel",
]

VT = t.TypeVar("VT")


class Inapplicable(BaseException):
    """Used to indicate that the source can not be used"""


class ConstantSource(enum.Enum):
    """Enumeration of constant effective value sources"""

    WORKFLOW = "workflow configuration"
    COMMAND = "CLI option"
    ENVIRONMENT = "environment variable"
    CONFIG = "configuration file"
    DEFAULT = "default value"
    MULTIPLE = "multiple sources"


@dataclasses.dataclass
class ConstantProxyDescriptor:
    """Descriptor for constant values"""

    name: str
    value: t.Any
    definition: t.Any
    effective_source: ConstantSource

    def __call__(self) -> t.Any:
        return self.value


class ConstantSentinelType(str):
    """Sentinel value type for ConstantBase"""


sentinel = ConstantSentinelType()


# pylint: disable=missing-function-docstring
class ConstantBase(WithLogger, t.Generic[VT]):
    """Constants used in grana runtime"""

    ENVIRONMENT_VARIABLE_NAME: str = sentinel
    COMMAND_LINE_OPTION_NAME: str = sentinel
    RC_PARAMETER_NAME: str = sentinel
    WORKFLOW_CONFIG_PARAMETER_NAME: str = sentinel
    DEFAULT: t.Union[VT, ConstantSentinelType] = sentinel

    def __init__(self) -> None:
        self._name: str = ""
        self._result_and_source: t.Union[tuple[VT, ConstantSource], ConstantSentinelType] = sentinel

    def __set_name__(self, owner: type, name: str) -> None:
        self._name = name

    def _register_result(self, result: VT, source: ConstantSource) -> None:
        if self._result_and_source is sentinel:
            self.logger.debug(f"{self._name} is accepted from {source}")
            self._result_and_source = result, source

    def __get__(self, instance: t.Any, owner: type) -> VT:
        return self.get()

    @CACHE.wrap
    def get(self) -> VT:
        """To be cached in the context"""
        self._result_and_source = sentinel
        for source, method in (
            (ConstantSource.WORKFLOW, self.from_workflow_configuration),
            (ConstantSource.COMMAND, self.from_cli_option),
            (ConstantSource.ENVIRONMENT, self.from_env),
            (ConstantSource.CONFIG, self.from_rc_file),
            (ConstantSource.DEFAULT, self.default),
        ):
            try:
                result = method()
            except Inapplicable:
                continue
            except Exception as e:
                self.logger.warning(f"Evaluating {self._name!r} from {source.value} failed: {e!r}")
                raise
            self._register_result(result=result, source=source)
        if isinstance(self._result_and_source, ConstantSentinelType):
            raise NotImplementedError
        effective_result, effective_source = self._result_and_source
        self.logger.info(f"Effective value for {self._name!r} is {effective_result!r} (from {effective_source.value})")
        return t.cast(
            VT,
            LazyProxy(
                ConstantProxyDescriptor(
                    name=self._name,
                    value=effective_result,
                    definition=self,
                    effective_source=effective_source,
                )
            ),
        )

    def cast(self, value: t.Any) -> VT:
        """Transform a value into the desired type"""
        return t.cast(VT, value)

    def from_workflow_configuration(self) -> VT:
        # Try to load the value from loaded workflow configuration variables
        wf_cfg_param_name: str = self.WORKFLOW_CONFIG_PARAMETER_NAME
        if wf_cfg_param_name is sentinel:
            raise Inapplicable
        ctx_values_map: dict[str, t.Any] = workflow.CONTEXT_HOLDER.get()
        if wf_cfg_param_name not in ctx_values_map:
            raise Inapplicable
        workflow_config_parameter_value: t.Any = ctx_values_map[wf_cfg_param_name]
        self.logger.debug(f"Defined workflow configuration value {wf_cfg_param_name!r} is accessed by {self._name!r}")
        return self.cast(workflow_config_parameter_value)

    def from_cli_option(self) -> VT:
        # Try to load the value from CLI options
        cli_option_name: str = self.COMMAND_LINE_OPTION_NAME
        if cli_option_name is sentinel:
            raise Inapplicable
        if (cli_option_value := get_cli_option(cli_option_name)) is None:
            raise Inapplicable
        self.logger.debug(f"Defined CLI option {cli_option_name!r} is accessed by {self._name!r}")
        return self.cast(cli_option_value)

    def from_env(self) -> VT:
        # Try to load the value from environment variables
        env_var_name: str = self.ENVIRONMENT_VARIABLE_NAME
        if env_var_name is sentinel:
            raise Inapplicable
        if (env_var_value := os.environ.get(env_var_name)) is None:
            raise Inapplicable
        self.logger.debug(f"Defined environment variable {env_var_name!r} is accessed by {self._name!r}")
        return self.cast(env_var_value)

    def from_rc_file(self) -> VT:
        # Try to load the value from the RC file
        rc_param_name: str = self.RC_PARAMETER_NAME
        if rc_param_name is sentinel:
            raise Inapplicable
        cfg: rc.RC = rc.RC.build()
        rc_param_value: t.Any = getattr(cfg, rc_param_name)
        if rc_param_value is rc.sentinel:
            raise Inapplicable
        return self.cast(rc_param_value)

    def default(self) -> VT:
        # Default value to be applied after every other source has been tested
        if isinstance(self.DEFAULT, ConstantSentinelType):
            raise Inapplicable  # pragma: no cover
        return self.DEFAULT


class ConstantBool(ConstantBase[bool]):
    """Base for boolean constants"""

    def cast(self, value: t.Union[str, bool]) -> bool:
        if isinstance(value, bool):
            return value
        if value == "Y":
            return True
        if value == "N":
            return False
        if value == "":
            raise Inapplicable
        raise ValueError(f"{value!r} is not a valid value for a boolean variable. Expected one of: 'Y', 'N'.")


class ConstantPathList(ConstantBase[list[Path]]):
    """Base for path list constants"""

    def _register_result(self, result: list[Path], source: ConstantSource) -> None:
        """Cumulative constant processing"""
        if self._result_and_source is not sentinel:
            prev_result, _ = self._result_and_source
            result += prev_result  # type: ignore[arg-type]
            source = ConstantSource.MULTIPLE
        self._result_and_source = result, source

    def cast(self, value: t.Union[str, list[Path]]) -> list[Path]:
        if isinstance(value, str):
            return [Path(item.strip()) for item in value.split(":") if item]
        return value

    def default(self) -> list[Path]:
        """An empty list"""
        return []


class ConstantPath(ConstantBase, t.Generic[VT]):
    """Base class for path constants"""

    def cast(self, value: t.Union[str, Path]) -> Path:
        return Path(value)



"""Shell action wrapped into a docker container"""

import asyncio
import contextlib
import functools
import tempfile
import typing as t
import uuid
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path

import aiodocker
import aiohttp
from aiodocker.containers import DockerContainer
from aiohttp.client import DEFAULT_TIMEOUT

from ..base import EmissionScannerActionBase, ArgsBase
from ..types import Stderr
from ...config.constants import C

__all__ = [
    "DockerShellArgs",
    "DockerShellAction",
]


class BindMode(Enum):
    """Allowed bind mount modes"""

    READ_ONLY = "ro"
    READ_WRITE = "rw"


class NetworkMode(Enum):
    """Allowed network modes"""

    BRIDGE = "bridge"
    HOST = "host"
    NONE = "none"


@dataclass
class Auth:
    """Docker auth info"""

    username: str
    password: str
    hostname: t.Optional[str] = None


@dataclass
class Network:
    """Container network specification"""

    mode: NetworkMode = NetworkMode.BRIDGE


@dataclass
class FileDockerBind:
    """File-based bind mount specification"""

    src: str
    dest: str
    mode: BindMode = BindMode.READ_WRITE


@dataclass
class ContentDockerBind:
    """Content-based bind mount specification"""

    contents: str
    dest: str
    mode: BindMode = BindMode.READ_ONLY


class DockerShellArgs(ArgsBase):
    """Args for docker shell"""

    command: str
    image: str
    environment: t.Optional[dict[str, str]] = None
    cwd: t.Optional[str] = None
    pull: bool = False
    executable: str = "/bin/sh"
    bind: t.Optional[list[t.Union[FileDockerBind, ContentDockerBind]]] = None
    network: Network = field(default_factory=Network)  # pylint: disable=invalid-field-call
    privileged: bool = False
    auth: t.Optional[Auth] = None


class DockerShellAction(EmissionScannerActionBase):
    """Runs a shell command in a docker container."""

    args: DockerShellArgs
    _ENTRY_SCRIPT_FILE_NAME: str = "entry.sh"
    _CONTAINER_TMP_DIRECTORY: str = "/tmp-grana"  # nosec

    @contextlib.asynccontextmanager
    async def _make_container(self, client: aiodocker.Docker) -> t.AsyncGenerator[DockerContainer, None]:
        container_name = f"grana-docker-shell-{uuid.uuid4().hex}"
        self.logger.info(f"Starting docker shell container {container_name!r}")
        container_entry_file_path: str = f"{self._CONTAINER_TMP_DIRECTORY}/{self._ENTRY_SCRIPT_FILE_NAME}"
        with tempfile.TemporaryDirectory() as tmp_directory:
            local_tmp_dir_path: Path = Path(tmp_directory)
            local_tmp_dir_path.chmod(0o777)
            self.logger.debug(f"Local temp dir is {local_tmp_dir_path}")
            entry_file_content: str = self.args.command
            if C.SHELL_INJECT_YIELD_FUNCTION:
                entry_file_content = f"{self._SHELL_SERVICE_FUNCTIONS_DEFINITIONS}\n{entry_file_content}"
            bind_configs: list[t.Union[FileDockerBind, ContentDockerBind]] = [
                ContentDockerBind(
                    contents=entry_file_content,
                    dest=container_entry_file_path,
                    mode=BindMode.READ_WRITE,
                )
            ]
            if self.args.bind:
                bind_configs += self.args.bind
            container_binds: list[str] = []
            for bind_config in bind_configs:
                if isinstance(bind_config, FileDockerBind):
                    local_file_full_name: str = bind_config.src
                else:
                    bind_contents_local_file: Path = local_tmp_dir_path / uuid.uuid4().hex
                    bind_contents_local_file.write_text(data=bind_config.contents, encoding="utf-8")
                    bind_contents_local_file.chmod(0o777)
                    local_file_full_name = str(bind_contents_local_file)
                container_binds.append(f"{local_file_full_name}:{bind_config.dest}:{bind_config.mode.value}")
            self.logger.debug(f"Container volumes: {container_binds}")
            container: DockerContainer = await client.containers.run(
                name=container_name,
                config={
                    "Cmd": [self.args.executable, container_entry_file_path],
                    "Image": self.args.image,
                    "HostConfig": {
                        "Binds": container_binds,
                        "Init": True,
                        "NetworkMode": self.args.network.mode.value,
                    },
                    "Env": [f"{k}={v}" for k, v in (self.args.environment or {}).items()],
                    "WorkingDir": self.args.cwd,
                    "Privileged": self.args.privileged,
                },
                auth=self._make_auth(),
            )
            try:
                yield container
            finally:
                await container.delete(force=True)

    @functools.cache
    def _make_auth(self) -> t.Optional[dict[str, str]]:
        if self.args.auth is None:
            return None
        auth_dict: dict[str, str] = {
            "username": self.args.auth.username,
            "password": self.args.auth.password,
        }
        if self.args.auth.hostname is not None:
            auth_dict["serveraddress"] = self.args.auth.hostname
        return auth_dict

    async def run(self) -> None:
        async with aiodocker.Docker() as client:
            # Enable default timeout for the connect phase only
            # pylint: disable=protected-access
            client.session._timeout = aiohttp.ClientTimeout(
                connect=DEFAULT_TIMEOUT.total,
            )
            if self.args.pull:
                self.logger.info(f"Pulling image: {self.args.image!r}")
                await client.pull(
                    from_image=self.args.image,
                    auth=self._make_auth(),
                )
            async with self._make_container(client) as container:
                tasks: list[asyncio.Task] = [
                    asyncio.create_task(self._read_stdout(container)),
                    asyncio.create_task(self._read_stderr(container)),
                ]
                # Wait for all tasks to complete
                await asyncio.wait(tasks)
                # Check exceptions
                await asyncio.gather(*tasks)
                result: dict = await container.wait()
                self.logger.debug(f"Docker container result: {result}")
                if (code := result.get("StatusCode", -1)) != 0:
                    self.fail(f"Exit code: {code}")

    async def _read_stdout(self, container: DockerContainer) -> None:
        async for chunk in container.log(stdout=True, follow=True):
            self.say(chunk)

    async def _read_stderr(self, container: DockerContainer) -> None:
        async for chunk in container.log(stderr=True, follow=True):
            self.say(Stderr(chunk))



"""Types collection"""

import abc
import dataclasses
import enum
import typing as t

__all__ = [
    "NamedMessageSource",
    "RenamedMessageSource",
    "Stderr",
    "Expression",
    "qualify_string_as_potentially_renderable",
    "ActionStatus",
]


class ActionStatus(enum.Enum):
    """Action valid states"""

    PENDING = "PENDING"  # Enabled, but not started yet
    RUNNING = "RUNNING"  # Execution in process
    SUCCESS = "SUCCESS"  # Finished without errors
    WARNING = "WARNING"  # Erroneous action with low severity
    FAILURE = "FAILURE"  # Erroneous action
    SKIPPED = "SKIPPED"  # May be set by action itself
    OMITTED = "OMITTED"  # Disabled during interaction

    def __repr__(self) -> str:
        return self.name

    __str__ = __repr__


class NamedMessageSource(t.Protocol, t.Hashable):
    """Anything that can be a message source (e.g. an action instance)"""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Source name"""

    @property
    @abc.abstractmethod
    def status(self) -> ActionStatus:
        """Source status"""

    @property
    @abc.abstractmethod
    def description(self) -> t.Optional[str]:
        """Source info"""


class RenamedMessageSource:
    """Renamed message source"""

    def __init__(self, name: str, origin: NamedMessageSource) -> None:
        self.name: str = name
        self._origin: NamedMessageSource = origin

    @property
    def status(self) -> ActionStatus:
        """Proxy to the origin status"""
        return self._origin.status

    @property
    def description(self) -> t.Optional[str]:
        """Proxy to the origin description"""
        return self._origin.description  # pragma: no cover


class Stderr(str):
    """Strings related to standard error stream"""


@dataclasses.dataclass
class Expression:
    """Complex object expression to be rendered later"""

    expression: str


def qualify_string_as_potentially_renderable(data: str) -> bool:
    """Check that a string should be templated later"""
    return "@{" in data



"""CLI options"""

import functools
import typing as t

import click

__all__ = [
    "cli_opts_receiver",
    "get_cli_option",
]

_CLI_OPTIONS: dict[str, t.Any] = {}


def cli_opts_receiver(func):
    """Store CLI options in the _CLI_OPTIONS container for further processing"""

    @functools.wraps(func)
    # pylint: disable=unused-argument
    def wrapped(ctx: click.Context, **kwargs):
        old_cli_options: dict[str, t.Any] = _CLI_OPTIONS.copy()
        current_ctx: t.Optional[click.Context] = ctx
        while current_ctx:
            for k, v in current_ctx.params.items():
                if k not in _CLI_OPTIONS:
                    _CLI_OPTIONS[k] = v
            current_ctx = current_ctx.parent
        try:
            return func()
        finally:
            # Restore CLI params container
            for k in list(_CLI_OPTIONS):
                del _CLI_OPTIONS[k]
            _CLI_OPTIONS.update(old_cli_options)

    return click.pass_context(wrapped)


def get_cli_option(name: str) -> t.Any:
    """Obtain previously registered CLI option"""
    return _CLI_OPTIONS.get(name)



"""Templar containers for rendering."""

import typing as t

from .proxy import LazyProxy
from ..exceptions import ActionRenderError, PendingActionUnresolvedOutcomeError

__all__ = [
    "AttrDict",
    "LooseDict",
    "OutcomeDict",
    "ActionContainingDict",
    "LazyProxy",
]

RenderHookType = t.Callable[[str], str]


class ItemAttributeAccessorMixin:
    """Anything, that can be accessed fie __getitem__, is available also as an attribute"""

    def __getattr__(self, item: str):
        return self.__getitem__(item)


class AttrDict(dict, ItemAttributeAccessorMixin):
    """Basic dictionary that allows attribute read access to its keys"""


class LooseDict(AttrDict):
    """A dictionary that allows attribute read access to its keys with a default empty value fallback"""

    def __getitem__(self, item: str):
        try:
            return super().__getitem__(item)
        except KeyError:
            return ""


class OutcomeDict(AttrDict):
    """A dictionary that allows attribute read access to its keys with a default value fallback"""

    def __getitem__(self, item: str):
        try:
            return super().__getitem__(item)
        except KeyError as e:
            from ..config.constants import C

            if C.STRICT_OUTCOMES_RENDERING:
                raise ActionRenderError(f"Outcome key {e} not found") from e
            return ""


class ActionContainingDict(AttrDict):
    """Anything with action names as keys"""

    def __getitem__(self, item: str):
        try:
            return super().__getitem__(item)
        except KeyError as e:
            raise ActionRenderError(f"Action not found: {e}") from e


class ActionOutcomeAggregateDict(ActionContainingDict):
    """Top-level container for action outcomes"""

    def __getitem__(self, item: str):
        if (result := super().__getitem__(item)) is not None:
            return result
        raise PendingActionUnresolvedOutcomeError(item)



"""Everything related to a default action interpretation"""

from __future__ import annotations

import asyncio
import base64
import collections
import copy
import dataclasses
import enum
import functools
import re
import textwrap
import typing as t

from .constants import ACTION_RESERVED_FIELD_NAMES
from .types import Stderr, ActionStatus, RenamedMessageSource, NamedMessageSource
from ..display.types import DisplayEvent, DisplayEventName
from ..exceptions import ActionRunError, ActionRenderError, ActionArgumentsLoadError
from ..logging import WithLogger, context
from ..rendering import WorkflowTemplar
from ..tools import classloader
from ..tools.classloader.exceptions import TypeMatchError, ClassLoaderError
from ..tools.inspect import get_class_annotations

__all__ = [
    "ActionDependency",
    "ActionSeverity",
    "ActionBase",
    "WorkflowActionExecution",
    "ActionSkip",
    "ArgsBase",
    "EmissionScannerActionBase",
]


# pylint: disable=unused-argument
class AbstractExecutionCommunicator(WithLogger):
    """Communication shim between action and its execution unit"""

    def send_say(self, message: str) -> None:
        """Pass a message to the execution"""
        self.logger.warning("`say` did not take effect")

    def send_yield_outcome(self, key: str, value: t.Any) -> None:
        """Pass an outcome to the execution"""
        self.logger.warning("`yield_outcome` did not take effect")

    def resend_display_event(self, event: DisplayEvent) -> None:
        """Pass a display event to the execution"""
        self.logger.warning("`send_display_event` did not take effect")


class ActionSkip(BaseException):
    """Stop executing action"""


class ActionSeverity(enum.Enum):
    """Action severity"""

    LOW = "low"
    NORMAL = "normal"


def strict_default_factory() -> bool:
    """Get default strictness value"""
    from ..config.constants import C

    return C.DEPENDENCY_DEFAULT_STRICTNESS


@dataclasses.dataclass
class ActionDependency:
    """Dependency info holder"""

    name: str
    strict: bool = dataclasses.field(default_factory=strict_default_factory)


class ArgsMeta(type):
    """Metaclass for args containers that makes them all dataclasses"""

    def __new__(cls, name, bases, dct):
        sub_dataclass = dataclasses.dataclass(super().__new__(cls, name, bases, dct))
        reserved_names_collisions: set[str] = {
            f.name for f in dataclasses.fields(sub_dataclass)
        } & ACTION_RESERVED_FIELD_NAMES
        if reserved_names_collisions:
            raise TypeError(f"Reserved names found in {name!r} class definition: {sorted(reserved_names_collisions)}")
        return sub_dataclass


@dataclasses.dataclass
class ArgsBase(metaclass=ArgsMeta):
    """Default empty args holder.
    Should be subclassed and then added to the `args` annotation of any action class."""


class ActionBase(WithLogger):
    """Base class for all actions"""

    args: ArgsBase

    def __init__(self) -> None:
        self._communicator: AbstractExecutionCommunicator = AbstractExecutionCommunicator()

    def yield_outcome(self, key: str, value: t.Any) -> None:
        """Report outcome key"""
        self.logger.debug(f"Yielding a key: {key!r}")
        self._communicator.send_yield_outcome(key, value)

    def say(self, message: str) -> None:
        """Send a message to the display"""
        self._communicator.send_say(message)

    def skip(self) -> t.NoReturn:
        """Set status to SKIPPED and stop execution"""
        raise ActionSkip

    def fail(self, message: str = "") -> t.NoReturn:
        """Set corresponding error message and raise an exception"""
        raise ActionRunError(message)

    async def run(self) -> None:
        """Main entry to be implemented in subclasses"""
        raise NotImplementedError


@dataclasses.dataclass
class WorkflowActionExecution(WithLogger):
    """An action that is executed within a workflow"""

    action_class: type[ActionBase]
    name: str
    raw_args: dict
    ancestors: list[ActionDependency] = dataclasses.field(default_factory=list)
    description: t.Optional[str] = None
    selectable: bool = True
    severity: ActionSeverity = ActionSeverity.NORMAL
    templar_factory: t.Optional[t.Callable[[dict], WorkflowTemplar]] = None
    locals_map: dict[str, t.Any] = dataclasses.field(default_factory=dict)

    def __post_init__(self) -> None:
        self.args_class: type[ArgsBase] = ArgsBase
        self.status: ActionStatus = ActionStatus.PENDING
        self.outcomes: dict[str, t.Any] = {}
        self.enabled: bool = True
        self.future: asyncio.Future = asyncio.get_event_loop().create_future()
        self.event_queue: asyncio.Queue[DisplayEvent] = asyncio.Queue()
        self._check_action_class_args()

    def __hash__(self) -> int:
        return id(self)

    def _check_action_class_args(self):
        """Validate action class `args` annotation
        and try the simplest loading of the dataclass from the original args map"""
        for mro_class in self.action_class.__mro__:
            if args_class := get_class_annotations(mro_class).get("args"):
                break
        else:
            raise ActionArgumentsLoadError(f"Couldn't find an `args` annotation for class {self.action_class.__name__}")
        try:
            self.args_class = classloader.get_data_class_by_data_signature(
                data_type=args_class,
                data=self.raw_args,
            )
        except (ValueError, ClassLoaderError) as e:
            raise ActionArgumentsLoadError(f"Action {self.name!r}: {e}") from e

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name!r}, status={self.status.value})"

    @functools.cache
    def compose_nested_source(self, origin: NamedMessageSource) -> NamedMessageSource:
        """Make a nested event"""
        return RenamedMessageSource(name=f"{self.name}/{origin.name}", origin=origin)

    async def _run_with_log_context(self) -> None:
        self.logger.info(f"Running action: {self.name!r}")
        execution = self

        class Communicator(AbstractExecutionCommunicator):
            """Closure-based communication interface"""

            def resend_display_event(self, event: DisplayEvent) -> None:
                new_event = DisplayEvent(name=event.name, **event.kwargs)
                new_event.future.add_done_callback(lambda _: event.future.set_result(None))
                if event.name == DisplayEventName.ON_RUNNER_START:
                    new_event.kwargs["children"] = map(execution.compose_nested_source, event.kwargs["children"])
                elif event.name in (
                    DisplayEventName.ON_ACTION_START,
                    DisplayEventName.ON_ACTION_FINISH,
                    DisplayEventName.ON_ACTION_MESSAGE,
                    DisplayEventName.ON_ACTION_ERROR,
                ):
                    new_event.kwargs["source"] = execution.compose_nested_source(event.kwargs["source"])
                elif event.name not in (
                    DisplayEventName.ON_RUNNER_FINISH,
                    DisplayEventName.ON_PLAN_INTERACTION,
                ):  # pragma: no cover
                    # Just in case we add some event types later and not specify behaviour here
                    raise ValueError(f"Unknown event name: {event.name!r}")  # pragma: no cover
                execution.event_queue.put_nowait(new_event)

            def send_say(self, message: str) -> None:
                execution.event_queue.put_nowait(
                    DisplayEvent(
                        DisplayEventName.ON_ACTION_MESSAGE,
                        source=execution,
                        message=message,
                    )
                )

            def send_yield_outcome(self, key: str, value: t.Any) -> None:
                execution.outcomes[key] = value

        action_instance: ActionBase = self.action_class()
        action_instance._communicator = Communicator()  # pylint: disable=protected-access
        with context(action=self.name):
            # Inject args
            action_instance.args = self.render_action_args()
            return await action_instance.run()

    def render_action_args(self) -> ArgsBase:
        """Prepare action to execution by rendering its template fields"""
        if self.templar_factory is None:
            return ArgsBase()

        templar = self.templar_factory(self.locals_map)
        fields: t.Dict[str, dataclasses.Field] = {f.name: f for f in dataclasses.fields(self.args_class)}
        rendered_args_dict: dict = {}
        for arg_key, arg_value in self.raw_args.items():
            corr_field: dataclasses.Field = fields[arg_key]
            if corr_field.metadata.get("rendering") == "disabled":
                self.logger.debug(f"Argument {arg_key!r} will not be rendered")
                rendered_args_dict[arg_key] = copy.deepcopy(arg_value)
            else:
                rendered_args_dict[arg_key] = templar.render(arg_value)
        try:
            parsed_args: ArgsBase = classloader.from_dict(
                data_type=self.args_class,
                data=rendered_args_dict,
            )
        except TypeMatchError as e:
            raise ActionRenderError(e) from None
        return parsed_args

    async def execute(self) -> None:
        """Wraps a call for the underlying action `run` method"""
        self.status = ActionStatus.RUNNING
        try:
            run_result = await self._run_with_log_context()  # type: ignore[func-returns-value]
        except ActionSkip:
            self.skip()
        except Exception as e:
            self.status = ActionStatus.FAILURE if self.severity == ActionSeverity.NORMAL else ActionStatus.WARNING
            self.logger.info(f"Action {self.name!r} failed: {repr(e)}")
            self.future.set_result(False)
            raise
        else:
            if run_result is not None:
                self.logger.warning(f"Action {self.name!r} return type is {type(run_result)} (not NoneType)")
            self.status = ActionStatus.SUCCESS
            self.future.set_result(True)

    def skip(self) -> None:
        """Skipping the action properly"""
        self.status = ActionStatus.SKIPPED
        self.future.set_result(True)
        self.logger.info(f"Action {self.name!r} skipped")

    def omit(self) -> None:
        """Omitting the action properly"""
        self.status = ActionStatus.OMITTED
        self.future.set_result(True)
        self.logger.info(f"Action {self.name!r} omitted")

    async def read_messages(self) -> t.AsyncGenerator[DisplayEvent, None]:
        """Obtain all said messages sequentially"""
        while True:
            # Wait for either an event or action finish
            queue_getter = asyncio.create_task(self.event_queue.get())
            await asyncio.wait(
                [self.future, queue_getter],
                return_when=asyncio.FIRST_COMPLETED,
            )
            if queue_getter.done():
                yield queue_getter.result()
            if self.future.done():
                # The action is done, so we should drain the queue.
                # Prevent queue from async get since then.
                queue_getter.cancel()
                while True:
                    try:
                        yield self.event_queue.get_nowait()
                    except asyncio.QueueEmpty:
                        break
                return


# pylint: disable=abstract-method
class EmissionScannerActionBase(ActionBase):
    """Base class for stream-scanning actions"""

    _SERVICE_MESSAGES_SCAN_PATTERN: t.ClassVar[t.Pattern] = re.compile(
        r"""^
          (.*?)  # possible preceding content
          \#\#grana\[  # message prefix
            ([A-Za-z0-9+/=\- ]+)  # message itself
          ]\#\#  # message suffix
        $""",
        re.VERBOSE,
    )
    _SHELL_SERVICE_FUNCTIONS_DEFINITIONS: str = textwrap.dedent(
        r"""
            yield_outcome()(
              [ "$1" = "" ] && echo "Missing key (first argument)" && return 1
              [ "$3" != "" ] && echo "Too many arguments (expected 1 or 2)" && return 2
              command -v base64 >/dev/null || ( echo "Missing command: base64" && return 3 )
              _pipe()(
                encodedKey="$1"
                while read -r data; do
                  if [ "$data" != "" ]; then
                    echo "##grana[yield-outcome-b64-chunk $encodedKey $data]##"
                  fi
                done
                echo "##grana[yield-outcome-b64-end $encodedKey]##"
              )
              encodedKey=$( printf "%s" "$1" | base64 | tr -d '\n' )
              if [ "$2" = "" ]; then
                base64 </dev/stdin | _pipe "$encodedKey"
              else
                printf "%s" "$2" | base64 | _pipe "$encodedKey"
              fi
              return 0
            )
            skip(){
              echo "##grana[skip]##"
              exit 0
            }
        """
    ).lstrip()

    def __init__(self) -> None:
        super().__init__()
        self._outcomes_base64_chunks: dict[str, list[str]] = collections.defaultdict(list)

    @classmethod
    def _decode_base64_string(cls, data: str) -> str:
        return base64.b64decode(data, validate=True).decode()

    def _process_service_message_expression(self, expression: str) -> None:
        try:
            expression_type, *encoded_args = expression.split()
            if expression_type == "skip":
                self.skip()
            elif expression_type == "yield-outcome-b64-chunk":
                key, value = encoded_args
                self._outcomes_base64_chunks[key].append(value)
            elif expression_type == "yield-outcome-b64-end":
                (encoded_key,) = encoded_args
                encoded_outcome_value: str = "".join(self._outcomes_base64_chunks.pop(encoded_key, []))
                self.yield_outcome(
                    key=self._decode_base64_string(encoded_key),
                    value=self._decode_base64_string(encoded_outcome_value),
                )
            else:
                raise ValueError(f"Unrecognized expression: {expression!r}")
        except ActionSkip:  # pylint: disable=try-except-raise
            raise
        except Exception:
            self.logger.warning("Failed while parsing system message", exc_info=True)

    def say(self, message: str) -> None:
        # Do not check stderr
        if isinstance(message, Stderr):
            super().say(message)
            return
        memorized_prefix: str = ""
        for line in message.splitlines():
            # `endswith` is a cheaper check than re.findall
            if not line.endswith("]##") or not (matches := self._SERVICE_MESSAGES_SCAN_PATTERN.findall(line)):
                super().say(memorized_prefix + line)
                memorized_prefix = ""
                continue
            for preceding_content, expression in matches:
                memorized_prefix += preceding_content
                self._process_service_message_expression(expression)
        # Do not forget to report system message prefix, if any
        if memorized_prefix:
            super().say(memorized_prefix)



"""Loader utilities"""

from __future__ import annotations

import pathlib

import yaml

from .context import LOADED_FILE_STACK
from ..actions.types import Expression
from ..exceptions import YAMLStructureError
from ..rendering import CommonTemplar

__all__ = [
    "ExpressionYAMLLoader",
    "DefaultYAMLLoader",
]


class ExpressionYAMLLoader(yaml.SafeLoader):
    """Extension parser"""

    @classmethod
    def add_string_constructor(cls, tag: str, target_class: type) -> None:
        """Register simple string constructor with type checking"""

        def construct(_, node):
            if not isinstance(node.value, str):
                raise YAMLStructureError(f"Expected string content after {tag!r}, got {node.value!r}")
            return target_class(node.value)

        cls.add_constructor(tag, construct)


ExpressionYAMLLoader.add_string_constructor("!@", Expression)


class DefaultYAMLLoader(ExpressionYAMLLoader):
    """Parser for default workflow loader"""

    @staticmethod
    def parse_load(loader: DefaultYAMLLoader, data: yaml.ScalarNode):
        """Process `!load` tag in parse-time to load external files"""
        templar: CommonTemplar = LOADED_FILE_STACK.create_associated_templar()
        file_path: pathlib.Path = pathlib.Path(templar.render(data.value))
        with file_path.open("rb") as f, LOADED_FILE_STACK.add(file_path):
            return yaml.load(f, loader.__class__)  # nosec


DefaultYAMLLoader.add_constructor("!load", DefaultYAMLLoader.parse_load)



"""Base interface class for all loaders"""

from __future__ import annotations

import collections
import contextlib
import dataclasses
import typing as t
from pathlib import Path

from .context import LOADED_FILE_STACK
from ..actions.base import WorkflowActionExecution, ActionBase, ActionDependency, ActionSeverity
from ..actions.constants import ACTION_RESERVED_FIELD_NAMES
from ..config.constants.workflow import WorkflowConfiguration
from ..exceptions import LoadError, ActionArgumentsLoadError
from ..logging import WithLogger
from ..rendering import WorkflowTemplar, CommonTemplar
from ..tools.classloader import from_dict
from ..workflow import Workflow

__all__ = [
    "AbstractBaseWorkflowLoader",
]


class AbstractBaseWorkflowLoader(WithLogger):
    """Loaders base class"""

    def __init__(self) -> None:
        self._executions: dict[str, WorkflowActionExecution] = {}
        self._gathered_context: dict[str, t.Any] = {}
        self._action_type_counters: dict[str, int] = collections.defaultdict(int)
        self._loaded_workflow: t.Optional[Workflow] = None
        self._loaded_config: WorkflowConfiguration = WorkflowConfiguration()

    @property
    def workflow(self) -> Workflow:
        """Loaded workflow getter"""
        # Check for typing and just in case
        if self._loaded_workflow is None:
            raise RuntimeError("No workflow was loaded")  # pragma: no cover
        return self._loaded_workflow

    @workflow.setter
    def workflow(self, workflow: Workflow) -> None:
        """Loaded workflow setter"""
        # Check for typing and just in case
        if self._loaded_workflow is not None:
            raise RuntimeError("Workflow was loaded already")  # pragma: no cover
        self._loaded_workflow = workflow

    def _register_action(self, action_execution: WorkflowActionExecution) -> None:
        if action_execution.name in self._executions:
            self._throw(f"Action declared twice: {action_execution.name!r}")
        self._executions[action_execution.name] = action_execution

    def _throw(self, message: str) -> t.NoReturn:
        """Raise loader exception from text"""
        raise LoadError(message=message, stack=LOADED_FILE_STACK.get_all()) from None

    def _internal_load_from_file(self, source_file: Path) -> None:
        """Load workflow partially from file (can be called recursively).
        :param source_file: either Path or string object pointing at a file"""
        with self._read_file(source_file) as file_data:
            self._internal_load_from_text(file_data)

    @contextlib.contextmanager
    def _read_file(self, source_file: Path) -> t.Iterator[str]:
        """Read file data"""
        source_resolved_file_path: Path = source_file.resolve()
        with LOADED_FILE_STACK.add(source_resolved_file_path):
            self.logger.debug(f"Loading workflow file: {source_resolved_file_path}")
            if not source_resolved_file_path.is_file():
                self._throw(f"Workflow file not found: {source_resolved_file_path}")
            yield source_resolved_file_path.read_text(encoding="utf-8")

    def _internal_load_from_text(self, data: str) -> None:
        """Load workflow partially from text (can be called recursively)"""
        raise NotImplementedError

    def _internal_load_from_dict(self, data: dict) -> None:
        """Load workflow partially from a dictionary (can be called recursively)"""
        raise NotImplementedError

    def get_action_factories_info(self) -> dict[str, tuple[type[ActionBase], str]]:
        """Returns a mapping of action factories names to its implementation classes and source information"""
        raise NotImplementedError

    def _get_action_factory_by_type(self, action_type: str) -> type[ActionBase]:
        action_info: t.Optional[tuple[type[ActionBase], str]] = self.get_action_factories_info().get(action_type)
        if action_info is None:
            self._throw(f"Unknown action type: {action_type}")
        return action_info[0]

    def load_from_text(self, data: str) -> Workflow:
        """Load workflow from text"""
        self._internal_load_from_text(data=data)
        self.workflow = Workflow(
            self._executions,
            context=self._gathered_context,
            configuration=self._loaded_config,
        )
        return self.workflow

    def load_from_file(self, source_file: Path) -> Workflow:
        """Load workflow from file"""
        self._internal_load_from_file(source_file=source_file)
        self.workflow = Workflow(
            self._executions,
            context=self._gathered_context,
            source_file=Path(source_file),
            configuration=self._loaded_config,
        )
        return self.workflow

    def load_from_dict(self, data: dict) -> Workflow:
        """Load workflow from file"""
        self._internal_load_from_dict(data=data)
        self.workflow = Workflow(
            self._executions,
            context=self._gathered_context,
            configuration=self._loaded_config,
        )
        return self.workflow

    def build_dependency_from_node(self, dep_node: t.Union[str, dict]) -> ActionDependency:
        """Unified method to process transform dependency source data"""
        if isinstance(dep_node, str):
            return ActionDependency(name=dep_node)
        if not isinstance(dep_node, dict):
            self._throw(f"Unrecognized dependency node structure: {type(dep_node)!r} (expected a string or a dict)")
        unexpected_dep_keys: set[str] = set(dep_node) - {"name", "strict", "external"}
        if unexpected_dep_keys:
            self._throw(f"Unrecognized dependency node keys: {sorted(unexpected_dep_keys)}")
        # Dependency name
        if "name" not in dep_node:
            self._throw(f"Name not specified for the dependency: {sorted(dep_node.items())}")
        dep_name: str = dep_node["name"]
        if not isinstance(dep_name, str):
            self._throw(f"Unrecognized dependency name type: {type(dep_name)!r} (expected a string)")
        if not dep_name:
            self._throw("Empty dependency name met")
        # Dependency 'strict' attr
        if "strict" not in dep_node:
            return ActionDependency(name=dep_name)
        strict: bool = dep_node["strict"]
        if not isinstance(strict, bool):
            self._throw(f"Unrecognized 'strict' attribute type: {type(strict)!r} (expected boolean)")
        return ActionDependency(name=dep_name, strict=strict)

    def build_action_from_dict_data(self, node: dict[str, t.Any]) -> WorkflowActionExecution:
        """Process a dictionary representing an action"""
        # Split node data into service fields and args
        service_fields: dict[str, t.Any] = {}
        raw_args: dict[str, t.Any] = {}
        for key, value in node.items():
            if key in ACTION_RESERVED_FIELD_NAMES:
                service_fields[key] = value
            else:
                raw_args[key] = value
        # Action type
        if "type" not in service_fields:
            self._throw("'type' not specified for action")
        action_type: str = service_fields["type"]
        action_class: type[ActionBase] = self._get_action_factory_by_type(action_type)
        # Action name
        name: str
        if "name" in service_fields:
            name = service_fields["name"]
            if not isinstance(name, str):
                self._throw(f"Unexpected name type: {type(name)!r} (should be a string")
            if not name:
                self._throw("Action node name is empty")
        else:
            if (action_counter := self._action_type_counters[action_type]) > 0:
                auto_name_suffix: str = f"-{action_counter + 1}"
            else:
                auto_name_suffix = ""
            name = f"{action_type}{auto_name_suffix}"
        self._action_type_counters[action_type] += 1
        # Description
        description: t.Optional[str] = service_fields.get("description", None)
        if description is not None and not isinstance(description, str):
            self._throw(f"Unrecognized 'description' content type: {type(description)!r} (expected optional string)")
        # Dependencies
        deps_node: t.Union[str, list[t.Union[str, dict]]] = service_fields.get("expects", [])
        if not isinstance(deps_node, str) and not isinstance(deps_node, list):
            self._throw(f"Unrecognized 'expects' content type: {type(deps_node)!r} (expected a string or list)")
        if isinstance(deps_node, str):
            deps_node = [deps_node]
        dependencies: list[ActionDependency] = [self.build_dependency_from_node(dep_node) for dep_node in deps_node]
        # Selectable
        selectable: bool = service_fields.get("selectable", True)
        if not isinstance(selectable, bool):
            self._throw(f"Unrecognized 'selectable' content type: {type(selectable)!r} (expected a bool)")
        # Severity
        severity_str: str = service_fields.get("severity", ActionSeverity.NORMAL.value)
        if not isinstance(severity_str, str):
            self._throw(f"Unrecognized 'severity' content type: {type(severity_str)!r} (expected a string)")
        try:
            severity = ActionSeverity(severity_str)
        except ValueError:
            valid_severities: str = ", ".join(sorted(s.value for s in ActionSeverity))
            self._throw(f"Invalid severity: {severity_str!r} (expected one of: {valid_severities})")
        locals_map: dict[str, t.Any] = service_fields.get("locals", {})
        if not isinstance(locals_map, dict):
            self._throw(f"'locals' contents should be a dict (got {type(locals_map)!r})")
        for local_key in locals_map:
            if not isinstance(local_key, str):
                self._throw(f"'locals' keys should be strings (got {type(local_key)!r} for {local_key!r})")
        try:
            action_instance: WorkflowActionExecution = WorkflowActionExecution(
                name=name,
                action_class=action_class,
                raw_args=raw_args,
                description=description,
                ancestors=dependencies,
                selectable=selectable,
                severity=severity,
                locals_map=locals_map,
                templar_factory=self._get_workflow_templar,
            )
        except ActionArgumentsLoadError as e:
            self._throw(str(e))
        return action_instance

    def load_configuration_from_dict(self, configuration_dict: dict[str, t.Any]) -> None:
        """Process configuration dictionary"""
        if not isinstance(configuration_dict, dict):
            self._throw(f"'configuration' contents should be a dict (got {type(configuration_dict)!r})")
        allowed_cfg_keys: set[str] = {field.name for field in dataclasses.fields(WorkflowConfiguration)}
        for unrecognized_cfg_key in sorted(set(configuration_dict) - allowed_cfg_keys):
            self.logger.warning(f"Unrecognized configuration key: {unrecognized_cfg_key!r}")
            configuration_dict.pop(unrecognized_cfg_key)
        templar: CommonTemplar = LOADED_FILE_STACK.create_associated_templar()
        rendered_configuration_dict: dict[str, t.Any] = templar.render(configuration_dict)
        self._loaded_config = from_dict(data_type=WorkflowConfiguration, data=rendered_configuration_dict)

    def _get_workflow_templar(self, locals_map: dict) -> WorkflowTemplar:
        return self.workflow.get_templar(locals_map)



"""Lazy-loaded constants helpers"""

import functools
import hashlib
import sys
import types
import typing as t
from contextlib import contextmanager
from importlib.machinery import ModuleSpec
from importlib.util import (
    spec_from_file_location,
    module_from_spec,
)
from pathlib import Path
from types import ModuleType

from ...exceptions import SourceError

__all__ = [
    "class_from_module",
]

VT = t.TypeVar("VT")
GetterType = t.Callable[[], t.Optional[VT]]

EXTERNALS_MODULES_PACKAGE: str = "grana.external"


@contextmanager
def add_sys_paths(*paths: Path) -> t.Iterator[None]:
    """Temporarily add paths to sys.path"""
    normalized_paths: list[str] = [str(path.absolute().expanduser()) for path in paths]
    for path in normalized_paths:
        sys.path.insert(0, path)
    try:
        yield
    finally:
        for path in normalized_paths:
            sys.path.remove(path)


@functools.cache
def load_external_module(source: Path, submodule_name: t.Optional[str] = None) -> ModuleType:
    """Load an external module"""
    from ..constants import C

    if not source.is_file():
        raise SourceError(f"Missing source module: {source}")
    if submodule_name is None:
        submodule_name = hashlib.md5(str(source).encode()).hexdigest()  # nosec  # pragma: no cover
    module_name: str = f"{EXTERNALS_MODULES_PACKAGE}.{submodule_name}"
    module_spec: t.Optional[ModuleSpec] = spec_from_file_location(
        name=module_name,
        location=source,
    )
    if module_spec is None:
        raise SourceError(f"Can't read module spec from source: {source}")
    module: ModuleType = module_from_spec(module_spec)
    with add_sys_paths(*C.EXTERNAL_PYTHON_MODULES_PATHS):
        module_spec.loader.exec_module(module)  # type: ignore
    sys.modules[module_name] = module
    return module


def class_from_module(
    source_path: Path,
    class_name: str,
    submodule_name: t.Optional[str] = None,
) -> type:
    """Get a class from an external module, if given"""
    module: types.ModuleType = load_external_module(source_path, submodule_name)
    if not hasattr(module, class_name):
        raise AttributeError(f"External module contains no class {class_name!r} in {source_path!r}")
    return getattr(module, class_name)



"""Actions-related constants"""

__all__ = [
    "ACTION_RESERVED_FIELD_NAMES",
]

ACTION_RESERVED_FIELD_NAMES: set[str] = {
    "name",
    "type",
    "description",
    "expects",
    "selectable",
    "severity",
    "locals",
}



# pylint: disable=invalid-field-call
"""Separate module for subflow action"""

import dataclasses
import typing as t
from collections.abc import Mapping, MutableMapping
from pathlib import Path

from ..base import ArgsBase, ActionBase
from ...display.types import DisplayEvent, DisplayEventName
from ...exceptions import ExecutionFailed
from ...rendering.containers import OutcomeDict

__all__ = [
    "SubflowAction",
]


class SubflowArgsByPath(ArgsBase):
    """Subflow arguments with the file path."""

    path: Path
    extra_context: dict[str, t.Any] = dataclasses.field(default_factory=dict)


class SubflowArgsBySpec(ArgsBase):
    """Subflow arguments with the spec."""

    actions: list[dict[str, t.Any]] = dataclasses.field(metadata={"rendering": "disabled"})
    context: dict[str, t.Any] = dataclasses.field(default_factory=dict, metadata={"rendering": "disabled"})
    configuration: dict[str, t.Any] = dataclasses.field(default_factory=dict, metadata={"rendering": "disabled"})
    extra_context: dict[str, t.Any] = dataclasses.field(default_factory=dict)


class SubflowAction(ActionBase):
    """Executes an independent workflow and passes the display events to the original runner."""

    args: t.Union[SubflowArgsByPath, SubflowArgsBySpec]

    async def run(self) -> None:
        from ...config.constants import C
        from ...runner import Runner

        action: SubflowAction = self

        def _resend_event_via_action(event: DisplayEvent) -> None:
            # These events shall not pass to the parent runner
            if event.name in (
                DisplayEventName.ON_RUNNER_FINISH,  # Triggers final status output
                DisplayEventName.ON_PLAN_INTERACTION,  # Pauses the execution
            ):
                event.future.set_result(None)  # Unlock the execution and continue
            else:
                self._communicator.resend_display_event(event)  # Pass modified event

        class SubflowRunner(Runner):
            """A runner that intercepts and filters out events"""

            async def _process_display_events(self) -> None:
                while True:
                    event: DisplayEvent = await self._events_flow.get()
                    _resend_event_via_action(event)

            @classmethod
            def _deep_update_context(cls, receiver: dict[str, t.Any], source: Mapping, path: str) -> dict[str, t.Any]:
                """Apply changes to the context"""
                for source_key, source_value in source.items():
                    sub_path: str = f"{path}.{source_key}" if path else source_key
                    if source_key not in receiver:
                        cls.logger.debug(f"Adding context: {sub_path}")
                        receiver[source_key] = source_value
                    elif isinstance(source_value, Mapping) and isinstance(receiver[source_key], MutableMapping):
                        cls.logger.debug(f"Merging context: {sub_path}")
                        receiver[source_key] = cls._deep_update_context(receiver[source_key], source_value, sub_path)
                    else:
                        cls.logger.debug(f"Rewriting context: {sub_path}")
                        receiver[source_key] = source_value
                return receiver

            async def run_async(self) -> None:
                self.workflow.context = self._deep_update_context(
                    receiver=self.workflow.context,
                    source=action.args.extra_context,
                    path="",
                )
                try:
                    return await super().run_async()
                finally:
                    for sub_action in self.workflow.values():
                        action.yield_outcome(sub_action.name, OutcomeDict(sub_action.outcomes))

        # Cache [re]mount is required since the subflow may reconfigure some fields
        with C.mount_context_cache():
            source: t.Union[Path, dict]
            if isinstance(action.args, SubflowArgsByPath):
                source = action.args.path
            else:
                source = {
                    "actions": action.args.actions,
                    "context": action.args.context,
                    "configuration": action.args.configuration,
                }
            runner = SubflowRunner(source)
            try:
                await runner.run_async()
            except ExecutionFailed:
                self.fail()



"""Runtime configuration module"""

from __future__ import annotations

import dataclasses
import logging
import pathlib
import typing as t

import yaml

from .cache import CACHE
from ...loader.utils import ExpressionYAMLLoader
from ...tools.classloader import from_dict
from ...tools.proxy import DeferredCallsProxy

logger = DeferredCallsProxy(logging.getLogger(__name__))

__all__ = [
    "RC",
    "logger",
    "sentinel",
]


class ConfigSentinel:
    """Sentinel type for runtime configuration parameters"""


sentinel = ConfigSentinel()


@dataclasses.dataclass
class RC:
    """Runtime configuration"""

    log_level: t.Union[str, ConfigSentinel] = sentinel
    log_file: t.Union[str, ConfigSentinel] = sentinel
    workflow_file: t.Union[pathlib.Path, ConfigSentinel] = sentinel
    workflow_loader_source_file: t.Union[pathlib.Path, ConfigSentinel] = sentinel
    display_source_file: t.Union[pathlib.Path, ConfigSentinel] = sentinel
    display_name: t.Union[str, ConfigSentinel] = sentinel
    strategy: t.Union[str, ConfigSentinel] = sentinel
    force_color: t.Union[bool, ConfigSentinel] = sentinel
    default_shell_executable: t.Union[str, ConfigSentinel] = sentinel
    shell_inject_yield_function: t.Union[bool, ConfigSentinel] = sentinel
    strict_outcomes_rendering: t.Union[bool, ConfigSentinel] = sentinel
    action_classes_directories: t.Union[list[pathlib.Path], ConfigSentinel] = sentinel
    external_python_modules_paths: t.Union[list[pathlib.Path], ConfigSentinel] = sentinel
    strict: t.Union[bool, ConfigSentinel] = sentinel

    @classmethod
    @CACHE.wrap
    def build(cls) -> RC:
        """Load from file"""
        from . import C
        from ...rendering import CommonTemplar

        rc_file_path: pathlib.Path = C.RC_FILE
        if not rc_file_path.is_file():
            logger.debug(f"No RC file found at {str(rc_file_path)!r}")
            return RC()
        logger.info(f"Loading RC file: {str(rc_file_path)!r}")
        with rc_file_path.open() as f:
            config_data: dict = t.cast(dict, yaml.load(f, ExpressionYAMLLoader))  # nosec
        templar: CommonTemplar = CommonTemplar.from_source_file(rc_file_path)
        rendered_data: t.Dict[str, t.Any] = templar.render(config_data)
        return from_dict(data_type=RC, data=rendered_data)



"""Simple printer"""

from ..base import ArgsBase, ActionBase

__all__ = [
    "EchoAction",
    "EchoArgs",
]


class EchoArgs(ArgsBase):
    """Echo arguments"""

    message: str


class EchoAction(ActionBase):
    """Prints a message to the output."""

    args: EchoArgs

    async def run(self) -> None:
        self.say(self.args.message)



"""Separate module for shell-related action"""

import asyncio
import contextlib
import os
import pathlib
import shlex
import typing as t
from asyncio.streams import StreamReader
from asyncio.subprocess import create_subprocess_shell, Process  # noqa
from subprocess import PIPE  # nosec

from ..base import ArgsBase, EmissionScannerActionBase
from ..types import Stderr
from ...config.constants import C

__all__ = [
    "ShellAction",
]


class ShellArgsByCommand(ArgsBase):
    """Args for shell-related actions with a command provided"""

    command: str
    environment: t.Optional[dict[str, str]] = None
    cwd: t.Optional[str] = None
    executable: t.Optional[str] = None


class ShellArgsByFile(ArgsBase):
    """Args for shell-related actions with a file provided"""

    file: pathlib.Path
    environment: t.Optional[dict[str, str]] = None
    cwd: t.Optional[str] = None
    executable: t.Optional[str] = None


class ShellAction(EmissionScannerActionBase):
    """Runs a shell command on the local system."""

    _BYTES_LINE_SEPARATOR: bytes = os.linesep.encode()
    _ENCODING: str = "utf-8"
    args: t.Union[ShellArgsByCommand, ShellArgsByFile]

    @classmethod
    async def _read_stream(cls, stream: StreamReader, strip_linesep: bool = True) -> t.AsyncGenerator[str, None]:
        async for chunk in stream:  # type: bytes
            if strip_linesep:
                chunk = chunk.rstrip(cls._BYTES_LINE_SEPARATOR)
            yield chunk.decode(cls._ENCODING)

    async def _read_stdout(self, process: Process) -> None:
        if process.stdout is None:
            raise ValueError("Process standard output is not available")
        async for line in self._read_stream(process.stdout):
            self.say(line)

    async def _read_stderr(self, process: Process) -> None:
        if process.stderr is None:
            raise ValueError("Process standard output is not available")
        async for line in self._read_stream(process.stderr):
            self.say(Stderr(line))

    async def _create_process(self) -> Process:
        command: str
        if isinstance(self.args, ShellArgsByCommand):
            command = self.args.command
        else:
            command = f". {shlex.quote(str(self.args.file))}"
        if C.SHELL_INJECT_YIELD_FUNCTION:
            command = f"{self._SHELL_SERVICE_FUNCTIONS_DEFINITIONS}\n{command}"
        environment: t.Optional[dict[str, str]] = None
        if self.args.environment is not None:
            environment = os.environ.copy()
            environment.update(self.args.environment)
        process = await create_subprocess_shell(
            cmd=command,
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
            env=environment,
            cwd=self.args.cwd,
            executable=self.args.executable or C.DEFAULT_SHELL_EXECUTABLE,
        )
        return process

    @contextlib.asynccontextmanager
    async def _control_process_lifecycle(self):
        process = await self._create_process()
        yield process
        if process.returncode is None:
            process.kill()
        # Close communication anyway
        await process.communicate()
        for stream in (process.stdout, process.stderr, process.stdin):
            if stream is None:
                continue
            stream._transport.close()  # type: ignore[union-attr]  # pylint: disable=protected-access

    async def _transmit_process_standard_streams(self, process: Process) -> None:
        tasks: list[asyncio.Task] = [
            asyncio.create_task(self._read_stdout(process)),
            asyncio.create_task(self._read_stderr(process)),
        ]
        # Wait for all tasks to complete
        await asyncio.wait(tasks)
        # Check exceptions
        await asyncio.gather(*tasks)

    async def run(self) -> None:
        async with self._control_process_lifecycle() as process:
            await self._transmit_process_standard_streams(process)
            await process.communicate()
            if process.returncode:
                self.fail(f"Exit code: {process.returncode}")
